"""
Created on 2023-02-19

@author: wf
"""

import os
import sys
import traceback
import webbrowser
from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter

from meta.metamodel import Context
from meta.mw import SMWAccess
from meta.sidif2linkml import SiDIF2LinkML
from meta.uml import PlantUml
from meta.version import Version


class MetaModelCmd:
    """
    command line interface for metamodel handling
    """

    def __init__(self, debug: bool = False):
        """
        constructor

        Args:
            debug (bool): if True switch debugging on
        """
        self.debug = debug
        self.error = None
        self.errMsg = None
        self.context = None

    @classmethod
    def getArgParser(cls):
        """
        get my argument parser
        """
        program_shortdesc = Version.description
        program_license = """%s

          Created by %s on %s.
          Copyright 2022-2025 Wolfgang Fahl. All rights reserved.

          Licensed under the Apache License 2.0
          http://www.apache.org/licenses/LICENSE-2.0

          Distributed on an "AS IS" basis without warranties
          or conditions of any kind, either express or implied.

        USAGE
        """ % (
            program_shortdesc,
            Version.authors,
            str(__date__),
        )
        parser = ArgumentParser(
            description=program_license, formatter_class=RawDescriptionHelpFormatter
        )
        parser.add_argument(
            "--about",
            help="show about info [default: %(default)s]",
            action="store_true",
        )
        parser.add_argument(
            "--context",
            default="MetaModel",
            help="context to read [default: %(default)s]",
        )
        parser.add_argument("--copyright", help="copyright message for diagrams")
        parser.add_argument(
            "-d", "--debug", dest="debug", action="store_true", help="show debug info"
        )
        parser.add_argument(
            "-wd",
            "--doc_width",
            help="Maximum width of documentation (default: %(default)s)",
            type=int,
            default=40,
        )
        parser.add_argument("-i", "--input", help="input file")

        parser.add_argument(
            "-l", "--linkml", action="store_true", help="create linkml yaml file"
        )
        parser.add_argument("-t", "--title", help="the title of a diagram")
        parser.add_argument(
            "-u", "--uml", action="store_true", help="create plantuml diagram"
        )
        parser.add_argument(
            "-at",
            "--withAt",
            action="store_true",
            help="generate with @startuml/@enduml bracket",
        )
        parser.add_argument(
            "--wikiId",
            default="wiki",
            help="id of the wiki to generate for [default: %(default)s]",
        )
        return parser

    def readContext(self, args):
        """
        read the context from the given args

        Args:
            args (Args): command line arguments
        """
        if args.input:
            result_tuple = Context.fromSiDIF_input(args.input, debug=args.debug)
            self.context, self.error, self.errMsg = result_tuple
        elif args.wikiId and args.context:
            self.wikiId = args.wikiId
            self.smwAccess = SMWAccess(args.wikiId)
            self.mw_contexts = self.smwAccess.getMwContexts()
            self.mw_context = self.mw_contexts.get(args.context, None)
            self.context, self.error, self.errMsg = Context.fromWikiContext(
                self.mw_context, debug=args.debug
            )

    def hasError(self):
        if self.error is not None:
            print(f"reading Context failed:\n{self.errMsg}\n{self.error}")
        return self.error

    def genUml(self, args:Namespace) -> str:
        """
        generate uml for the given context with the given name from the given wiki

        Args:
           args (Namespace): the command line arguments

        Returns:
            str: the PlantUml
        """
        self.readContext(args)
        uml=None
        if not self.hasError():
            uml = PlantUml(
                title=args.title,
                copyRight=args.copyright,
                withAt=args.withAt,
                doc_width=args.doc_width,
            )
            uml.fromDIF(self.context.dif)
        return uml


    def genLinkML(self, args:Namespace) -> str:
        """
        generate linkML yaml for the given command line arguments

        Args:
           args (Namespace): the command line arguments

        Returns:
            str: the uml markup
        """
        self.readContext(args)
        if not self.hasError():
            sidif2LinkML = SiDIF2LinkML(self.context)
            yaml_text = sidif2LinkML.asYaml()
            return yaml_text
        else:
            return None


__version__ = Version.version
__date__ = Version.date
__updated__ = Version.updated


def main(argv=None):  # IGNORE:C0111
    """main program."""

    if argv is None:
        argv = sys.argv[1:]

    program_name = os.path.basename(__file__)
    program_version = f"v{__version__}"
    program_build_date = str(__updated__)
    program_version_message = f"{program_name} ({program_version},{program_build_date})"

    try:
        parser = MetaModelCmd.getArgParser()
        # Setup argument parser
        if len(argv) < 1:
            parser.print_usage()
            sys.exit(1)

        args = parser.parse_args(argv)
        mm_cmd = MetaModelCmd(debug=args.debug)
        if args.about:
            print(program_version_message)
            print(f"see {Version.doc_url}")
            webbrowser.open(Version.doc_url)
        if args.uml:
            uml = mm_cmd.genUml(args)
            print(uml)
        elif args.linkml:
            mm_cmd = MetaModelCmd(debug=args.debug)
            linkml_yaml = mm_cmd.genLinkML(args)
            print(linkml_yaml)

    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise (e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        if args.debug:
            print(traceback.format_exc())
        return 2


DEBUG = 1
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())



"""
Created on 2020-11-12

@author: wf
"""

import textwrap
from typing import Dict
from sidif.sidif import DataInterchange


class PlantUml(object):
    """
    Plant UML support
    """

    # redundant to skinparams in pylodstorage.uml
    skinparams = """
' BITPlan Corporate identity skin params
' Copyright (c) 2015-2025 BITPlan GmbH
' see http://wiki.bitplan.com/PlantUmlSkinParams#BITPlanCI
' skinparams generated by com.bitplan.restmodelmanager
skinparam note {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam component {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam package {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam usecase {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam activity {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam classAttribute {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam interface {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam class {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
skinparam object {
  BackGroundColor #FFFFFF
  FontSize 12
  ArrowColor #FF8000
  BorderColor #FF8000
  FontColor black
  FontName Technical
}
hide Circle
' end of skinparams '
"""

    def __init__(
        self,
        copyRight=None,
        title=None,
        debug=False,
        withSkin: bool = True,
        withAt: bool = False,
        doc_width: int = 40,
    ):
        """
        Constructor
        """
        self.debug = debug
        self.withSkin = withSkin
        self.withAt = withAt
        self.uml = ""
        self.title = title
        self.copyRight = copyRight
        self.doc_width = doc_width

    def __str__(self):
        return self.atIt(self.uml)

    def atIt(self, markup: str):
        """
        Args:
            markup (str): the markup
        Returns:
            str: markup with @startuml and @enduml
        """
        if not self.withAt:
            return markup
        at_markup = f"""@startuml
{markup}
@enduml"""
        return at_markup

    def multi_line_doc(self, doc: str, width: int) -> str:
        """
        Returns the given documentation as a multiline string with a limited length per line.
        Lines are broken at whitespace.

        Args:
            doc (str): the documentation to wrap
            width (int): The maximum length of each line.
        Returns:
            str: the  Multiline string.
        """
        text = "\n".join(textwrap.wrap(doc, width=width))
        return text

    def asUmlDict(self, dif):
        """
        return the given DataInterchange as an UML Dict
        """
        uml = {"packages": {}, "topiclinks": {}}
        classKey = None
        classes = {}
        for _i, triple in enumerate(dif.triples):
            if self.debug:
                print(triple)
            if triple.p == "isA":
                itkey = triple.s
                if triple.o == "Context":
                    packageKey = itkey
                    packages = uml["packages"]
                    packages[itkey] = {"classes": {}}
                    it = packages[itkey]
                elif triple.o == "TopicLink":
                    links = uml["topiclinks"]
                    links[itkey] = {}
                    it = links[itkey]
                elif triple.o == "Property":
                    propKey = itkey
                    if not classKey:
                        print(f"invalid property {propKey} declared out of class scope")
                        continue
                    properties = classes[classKey]["properties"]
                    properties[propKey] = {}
                    it = properties[propKey]
                else:
                    isA = triple.o
                    classKey = itkey
                    classes = packages[packageKey]["classes"]
                    clazz_dict={"isA": isA, "properties": {}}
                    classes[itkey] = clazz_dict
                    it = clazz_dict
            elif triple.p == "extends":
                # inheritance relation e.g. Computer extends Device
                special_class = triple.s # Computer
                general_class = triple.o # Device
                clazz_dict={"extends": general_class, "properties": {}}
                classes[special_class]= clazz_dict
                it=clazz_dict
                pass
            elif triple.o == "it":
                if triple.p == "addsTo":
                    # @TODO might note be redundant but
                    # declaring a default
                    # redundant forward declaration
                    pass
                elif triple.p == "context":
                    parentKey = triple.s
                    packages[parentKey]["classes"][classKey] = classes[classKey]
                    pass
                elif triple.p == "topic":
                    parentKey = triple.s
                    classes[parentKey]["properties"][propKey] = properties[propKey]
                    pass
                else:
                    it[triple.p] = triple.s
        return uml

    def fromDIF(self, dif:DataInterchange) -> str:
        """
        create a UML from a Data Interchange

        Args:
            dif (DataInterchange): - the data interchange

        Returns:
            str: the planuml markup

        """
        umlDict = self.asUmlDict(dif)
        self.uml = self.fromUmlDict(umlDict)
        return

    def fromUmlDict(self, umlDict: Dict) -> str:
        """
        convert the given umlDict Dict to a plantuml string

        Args:
            umlDict (Dict): the dictionary of packages,classes and properties

        Returns:
            str: the planuml markup
        """
        uml = ""
        if self.title is not None:
            if self.copyRight is None:
                copyRight = ""
            else:
                copyRight = "\n%s" % self.copyRight
            uml += f"title\n{self.title}{copyRight}\nend title\n"
        packages = umlDict["packages"]
        for packageKey in packages.keys():
            package = packages[packageKey]
            package_name=package.get('name',packageKey)
            uml += f"package {package_name} {{\n"
            for classKey in package["classes"]:
                uclass = package["classes"][classKey]
                className = uclass.get("name", classKey)
                extends_class = uclass.get("extends")
                if extends_class:
                    extends = f"extends {extends_class} "
                else:
                    extends = ""
                uml += f"  class {className} {extends}{{\n"
                for propKey in uclass["properties"]:
                    prop = uclass["properties"][propKey]
                    uml += f"    {prop['name']}:{prop['type']}\n"
                uml += "  }\n"
                if "documentation" in uclass:
                    doc = uclass["documentation"]
                    doc = self.multi_line_doc(doc, self.doc_width)
                    uml += f"""Note top of {className}
{doc}
End note
"""
            uml += "}\n"

        links = umlDict["topiclinks"]
        for linkKey in links.keys():
            link = links[linkKey]
            sourceMany = "*" if link["sourceMultiple"] else "1"
            targetMany = "*" if link["targetMultiple"] else "1"
            sourceRole = link["sourceRole"] if "sourceRole" in link else ""
            targetRole = link["targetRole"] if "targetRole" in link else ""
            source = link["source"]
            target = link["target"]
            uml += f'{source} "{sourceRole} {sourceMany}" -- "{targetRole} {targetMany}" {target}\n'

        if self.withSkin:
            uml += PlantUml.skinparams
        return uml



# Auto generated from wiki.yaml by pythongen.py version: 0.0.1
# Generation date: 2024-08-07T16:54:54
# Schema: WikiContext
#
# id: WikiContext
# description:
# license:

import dataclasses
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from datetime import date, datetime
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.metamodelcore import Bool, Curie, Decimal, ElementIdentifier, NCName, NodeIdentifier, URI, URIorCURIE, XSDDate, XSDDateTime, XSDTime

metamodel_version = "1.7.0"
version = None

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
WIKICONTEXT = CurieNamespace('WikiContext', 'WikiContext/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
SCHEMA = CurieNamespace('schema', 'http://schema.org/')
SHEX = CurieNamespace('shex', 'http://www.w3.org/ns/shex#')
XSD = CurieNamespace('xsd', 'http://www.w3.org/2001/XMLSchema#')
DEFAULT_ = WIKICONTEXT


# Types
class String(str):
    """ A character string """
    type_class_uri = XSD["string"]
    type_class_curie = "xsd:string"
    type_name = "string"
    type_model_uri = WIKICONTEXT.String


class Integer(int):
    """ An integer """
    type_class_uri = XSD["integer"]
    type_class_curie = "xsd:integer"
    type_name = "integer"
    type_model_uri = WIKICONTEXT.Integer


class Boolean(Bool):
    """ A binary (true or false) value """
    type_class_uri = XSD["boolean"]
    type_class_curie = "xsd:boolean"
    type_name = "boolean"
    type_model_uri = WIKICONTEXT.Boolean


class Float(float):
    """ A real number that conforms to the xsd:float specification """
    type_class_uri = XSD["float"]
    type_class_curie = "xsd:float"
    type_name = "float"
    type_model_uri = WIKICONTEXT.Float


class Double(float):
    """ A real number that conforms to the xsd:double specification """
    type_class_uri = XSD["double"]
    type_class_curie = "xsd:double"
    type_name = "double"
    type_model_uri = WIKICONTEXT.Double


class Decimal(Decimal):
    """ A real number with arbitrary precision that conforms to the xsd:decimal specification """
    type_class_uri = XSD["decimal"]
    type_class_curie = "xsd:decimal"
    type_name = "decimal"
    type_model_uri = WIKICONTEXT.Decimal


class Time(XSDTime):
    """ A time object represents a (local) time of day, independent of any particular day """
    type_class_uri = XSD["time"]
    type_class_curie = "xsd:time"
    type_name = "time"
    type_model_uri = WIKICONTEXT.Time


class Date(XSDDate):
    """ a date (year, month and day) in an idealized calendar """
    type_class_uri = XSD["date"]
    type_class_curie = "xsd:date"
    type_name = "date"
    type_model_uri = WIKICONTEXT.Date


class Datetime(XSDDateTime):
    """ The combination of a date and time """
    type_class_uri = XSD["dateTime"]
    type_class_curie = "xsd:dateTime"
    type_name = "datetime"
    type_model_uri = WIKICONTEXT.Datetime


class DateOrDatetime(str):
    """ Either a date or a datetime """
    type_class_uri = LINKML["DateOrDatetime"]
    type_class_curie = "linkml:DateOrDatetime"
    type_name = "date_or_datetime"
    type_model_uri = WIKICONTEXT.DateOrDatetime


class Uriorcurie(URIorCURIE):
    """ a URI or a CURIE """
    type_class_uri = XSD["anyURI"]
    type_class_curie = "xsd:anyURI"
    type_name = "uriorcurie"
    type_model_uri = WIKICONTEXT.Uriorcurie


class Curie(Curie):
    """ a compact URI """
    type_class_uri = XSD["string"]
    type_class_curie = "xsd:string"
    type_name = "curie"
    type_model_uri = WIKICONTEXT.Curie


class Uri(URI):
    """ a complete URI """
    type_class_uri = XSD["anyURI"]
    type_class_curie = "xsd:anyURI"
    type_name = "uri"
    type_model_uri = WIKICONTEXT.Uri


class Ncname(NCName):
    """ Prefix part of CURIE """
    type_class_uri = XSD["string"]
    type_class_curie = "xsd:string"
    type_name = "ncname"
    type_model_uri = WIKICONTEXT.Ncname


class Objectidentifier(ElementIdentifier):
    """ A URI or CURIE that represents an object in the model. """
    type_class_uri = SHEX["iri"]
    type_class_curie = "shex:iri"
    type_name = "objectidentifier"
    type_model_uri = WIKICONTEXT.Objectidentifier


class Nodeidentifier(NodeIdentifier):
    """ A URI, CURIE or BNODE that represents a node in a model. """
    type_class_uri = SHEX["nonLiteral"]
    type_class_curie = "shex:nonLiteral"
    type_name = "nodeidentifier"
    type_model_uri = WIKICONTEXT.Nodeidentifier


class Jsonpointer(str):
    """ A string encoding a JSON Pointer. The value of the string MUST conform to JSON Point syntax and SHOULD dereference to a valid object within the current instance document when encoded in tree form. """
    type_class_uri = XSD["string"]
    type_class_curie = "xsd:string"
    type_name = "jsonpointer"
    type_model_uri = WIKICONTEXT.Jsonpointer


class Jsonpath(str):
    """ A string encoding a JSON Path. The value of the string MUST conform to JSON Point syntax and SHOULD dereference to zero or more valid objects within the current instance document when encoded in tree form. """
    type_class_uri = XSD["string"]
    type_class_curie = "xsd:string"
    type_name = "jsonpath"
    type_model_uri = WIKICONTEXT.Jsonpath


class Sparqlpath(str):
    """ A string encoding a SPARQL Property Path. The value of the string MUST conform to SPARQL syntax and SHOULD dereference to zero or more valid objects within the current instance document when encoded as RDF. """
    type_class_uri = XSD["string"]
    type_class_curie = "xsd:string"
    type_name = "sparqlpath"
    type_model_uri = WIKICONTEXT.Sparqlpath


# Class references




# Enumerations


# Slots
class slots:
    pass


